# encoding=utf-8
import jieba
import xlrd
import xlwt
import math
import re
import json
import datetime
import sys, getopt

from groupingsentences.ssdistance import get_similarity_val_by_sentences
from groupingsentences.ssdistance import get_funcname_by_func_type_id
from groupingsentences.fileload import load_cells_from_file

#1：词向量文本分类，这套程序的核心思路是：
#a: 两两比对词库里的关键词
#b: 比对时计算两者之间的余弦值
#c: 根据返回的余弦值选择是否归为一类（修改该关键词所处顺序）
#d: 输出排序后的结果放在文档里


def cells_to_groups(cells, type_of_func = 0 , threshold = 0.55):
    groups = {}
    sentences_processed = set()

    for sentence1 in cells:
        if sentence1 not in sentences_processed:
            sentence_related = set()
            groups[sentence1] = sentence_related
            sentences_processed.add(sentence1)
            for sentence2 in cells:
                #print(sentence1)
                if sentence2 not in sentences_processed:
                    dist1 = get_similarity_val_by_sentences(sentence1, sentence2, type_of_func)
                    if dist1 > threshold:
                        #related
                        sentence_related.add(sentence2)
                        sentences_processed.add(sentence2)
    return groups

def save_groups_to_xls(dest_file_name, groups):
    #纵向优先
    #写入xls文件
    # 打开文件
    style0 = xlwt.easyxf('font: name Times New Roman, color-index black, bold off',
        num_format_str='#,##0.00')
    stylered = xlwt.easyxf('font: name Times New Roman, color-index red, bold off',
        num_format_str='#,##0.00')

    style1 = xlwt.easyxf(num_format_str='D-MMM-YY')
    wb = xlwt.Workbook()
    ws = wb.add_sheet('Sheet0')
    max_j_start = 0
    max_j_height = 0
    sorted_word_groups = sorted(groups.items(), key=lambda x: len(x[1]), reverse = True)
    total_counts = 0
    for key, value in sorted_word_groups:
        total_counts = total_counts + len(value) + 1
    print('total counts', total_counts)
    max_height = (total_counts + len(sorted_word_groups) + 254)/255 # max height we need
    print('max_height', max_height)

    style_index = 0
    i = 0 #column
    j = 0 #row
    for key, value in sorted_word_groups: #sort by the len of set() 
        #print('## j,i row,column',j,i)
        style_index = style_index + 1
        style_item = style0
        if style_index%2 == 0:
            style_item = stylered
        ws.write(j, i, key , style_item)
        j = j + 1
        if j > max_height:
            j = 0
            i = i + 1
        for value2 in sorted(value):
            #print('j,i',j,i)
            ws.write(j, i, value2 , style_item)
            j = j + 1
            if j > max_height:
                j = 0
                i = i + 1
        j = j + 1 # add a space
        if j > max_height:
            j = 0
            i = i + 1
    wb.save(dest_file_name)

def gs_grouping_sentences_to_xls(inputfile, outputfile, max_items=10000, encoding='gb18030', type_of_func = 0, threshold = 0.55):
    cells = load_cells_from_file(inputfile, encoding , max_items)
    ts = datetime.datetime.now().timestamp() 
    groups = cells_to_groups(cells, type_of_func, threshold)
    print('groups items count', len(groups))
    ts2 = datetime.datetime.now().timestamp() 
    save_groups_to_xls(outputfile, groups)
    ts3 = datetime.datetime.now().timestamp() 
    print('type of compare function', type_of_func)
    print('total cells count', len(cells))
    print('compare sentences time passed (seconds):', ts2-ts)
    print('save to xls time passed (seconds):', ts3-ts2) 

def main(argv):
    inputfile = ''
    outputfile = ''
    type_of_func = 0
    threshold = 0.55
    encoding = 'gb18030'
    max_items = 1000

    try:
        opts, args = getopt.getopt(argv,"hi:o:c:e:f:t:",["ifile=","ofile=","encoding=","maxcount="])
    except getopt.GetoptError:
        print('first.py -i <inputfile> -o <outputfile> -c <maxcount 1000 default> -e <encoding gb18030 default> -f <typeofcomparefunction 0 default> -t <threshold 0.55 default>')
        sys.exit(2)
    for opt, arg in opts:
        if opt == '-h':
            print('first.py -i <inputfile> -o <outputfile> -c <maxcount 1000 default> -e <encoding gb18030 default> -f <typeofcomparefunction 0 default> -t <threshold 0.55 default>')
            sys.exit()
        elif opt in ("-i", "--ifile"):
            inputfile = arg
        elif opt in ("-o", "--ofile"):
            outputfile = arg
        elif opt in ("-f"):
            type_of_func = int(arg)
        elif opt in ("-t"):
            threshold = float(arg)
        elif opt in ("-e", "--encoding"):
            encoding = (arg)
        elif opt in ("-c", "--maxcount"):
            max_items = int(arg)
    if len(inputfile)==0 or len(outputfile)==0:
        print('first.py -i <inputfile> -o <outputfile> -c <maxcount 1000 default> -e <encoding gb18030 default> -f <typeofcomparefunction 0 default> -t <threshold 0.55 default>')
        sys.exit(2)

    print('Input file is "', inputfile)
    print('Output file is "', outputfile)
    gs_grouping_sentences_to_xls(inputfile, outputfile, max_items, encoding, type_of_func , threshold)
   

if __name__ == "__main__":
    main(sys.argv[1:])



from groupingsentences.ssdistance import get_similarity_val_by_sentences
from groupingsentences.ssdistance import get_funcname_by_func_type_id
from groupingsentences.fileload import load_cells_from_file
from groupingsentences.first import cells_to_groups
from groupingsentences.first import save_groups_to_xls
from groupingsentences.second import cells_to_xmind
from groupingsentences.second import gs_grouping_sentences_to_xmind
from groupingsentences.first import gs_grouping_sentences_to_xls



# -*- coding: utf-8 -*-
"""
:
"""

import os
import sys, getopt

def load_cells_from_file(file_name, encoding='gb18030', max_items = 500000):
    #FILENAME = './dataset/1.csv'
    if os.path.isfile(file_name) == False:
        print('file not exist:', file_name)
        sys.exit(2)
        return
    cells = []
    count = 0
    with open(file_name, 'r', encoding=encoding) as f:
        while True: 
            count += 1
            # Get next line from file 
            line = f.readline() 
            # if line is empty 
            # end of file is reached 
            if not line: 
                break
            if count > max_items:
                break
            #print("Line{}: {}".format(count, line.strip())) 
            arr = line.strip().split(',')
            if len(arr) > 0:
                cells.append(arr[0])
            else:
                cells.append(line.strip())
        f.close()
    return cells


# encoding=utf-8
import jieba
import xlrd
import math
import re
import xmind
import json
import datetime
from groupingsentences.ssdistance import get_similarity_val_by_sentences
from groupingsentences.ssdistance import get_funcname_by_func_type_id
from groupingsentences.fileload import load_cells_from_file
import sys, getopt
import os.path

#关键词根提取法，这套程序的核心思路是：

#a: 统计top词根N个
#b: 提取包含词根的对应关键词子集
#c: 利用xmind模块，在循环中插入对应节点
#21：top词根不一定是合适的 同时 已经被作为节点的词根不应该重复计算
#因此需要一个库，记录已经被作为节点的词根
#同时，类似”可是“这样的词不具有研究意义，不应该被作为节点，所以需要一个库，事先导入非法词根表
#22：几百万的关键词，当需要打印的层级较多时，提取各级词频对应词库的交集很耗时，而且分词操作在遍历的过程中可以预计是会重复执行的，因此还是一样要做分词预处理，只不过这个预处理会更复杂一点，需要建立几个词典库：
#库1：{id:[长尾词，对应词根集合]}
#如：记录id”1“，对应值是一个列表，里面有长尾词”QQ邮箱格式怎么写“，和这个长尾词对应的所有词根集合（set(["QQ","邮箱","格式","怎么","写"])）
#库2：{长尾词:id}
#如：记录”QQ邮箱格式怎么写“，它对应在库1的记录id是”1“
#库3：{词根:对应词频}
#如：记录”QQ“，它的词频是5，说明它出现在5个关键词里
#库4：{词根:对应ID集合}
#如：记录”QQ“，它的对应ID集合是set([1,2,3,4,5])，表示有包含它的关键词的对应ID是哪几个

def create_stores(cells, words_already, illeagle_words):
    value_id = 1
    store1 = {}
    store2 = {}
    store3 = {}
    store4 = {}

    for sentence1 in cells:
        seg_list = jieba.cut_for_search(sentence1)
        for seg in list(seg_list):
            if seg not in words_already:
                seg_count = store3.get(seg, 0)
                store3[seg] = seg_count+1
                if seg in store4:
                    store4[seg].add(value_id) 
                else:
                    store4[seg] = set([value_id])
        seg_list = jieba.cut_for_search(sentence1)
        #print(",".join(list(seg_list)))
        if value_id not in store1:
            store1[value_id]=[sentence1,set(list(seg_list))]
        if sentence1 not in store2:
            store2[sentence1]=value_id
        value_id = value_id+1
    return store1,store2,store3,store4

#TOP 获取lw句子列表里对应的首N个词根列表
def get_top(lw, top_count, store1, store2, store3, words_already):
    #lw is a sentence list
    top_words = []
    segs = {}
    for value in lw:
        seg_list = store1[store2[value]][1] #库1：{id:[长尾词，对应词根集合]}
        for seg in list(seg_list):
            if seg not in words_already:
                segs[seg]=store3[seg] #库3：{词根:对应词频}
    sorted_segs = sorted(segs.items(), key = lambda kv:(kv[1], kv[0]),reverse=True)
    for item in sorted_segs[:top_count]:
        top_words.append(item[0])
    return top_words

def get_first_topwords(top_words_count, words_already, store3):
    top_words_list=[]

    #print(sorted(store3.items(), key = lambda kv:(kv[1], kv[0]),reverse=True)) 
    sorted_ku3 = sorted(store3.items(), key = lambda kv:(kv[1], kv[0]),reverse=True)
    # TOP 8
    k = 0
    for item in sorted_ku3[:top_words_count*3]:
        if item[0] not in words_already:
            top_words_list.append(item[0])
            words_already.add(item[0])
        k = k+1
        if k>top_words_count:
            break;
    print('top_words_list items count', len(top_words_list))
    return top_words_list

def gen_my_xmind_file(top_words_list, dest_file_name, store1, store2, store3, store4, words_already, top_words_count=8):  
    # 1、如果指定的XMind文件存在，则加载，否则创建一个新的
    workbook = xmind.load("./my.xmind")
    # 新建一个画布
    #sheet = workbook.createSheet()
    primary_sheet = workbook.getPrimarySheet()
    root_topic = primary_sheet.getRootTopic()
    # 给中心主题添加一个星星图标
    #root_topic.addMarker(MarkerId.starRed)
    #design_sheet1(primary_sheet)
    primary_sheet.setTitle("ROOT") #.decode('utf-8')
    # 新建一个主题
    root_topic = primary_sheet.getRootTopic()
    root_topic.setTitle("ROOT") #.decode('utf-8')
    for word1 in top_words_list:
        word1_count = (store3[word1].__str__())
        sub_root_topic = root_topic.addSubTopic()
        sub_root_topic.setTitle(word1 + ' ('+ word1_count +')') #.decode('utf-8')
        # 获取词根所有id
        idsets = store4[word1] #库4：{词根:对应句子ID集合}
        # 提取词根所有长尾词
        lw = [store1[ids][0] for ids in idsets]  #{id:[长尾词，对应词根集合]}  lw = 长尾词 list
        # 统计子top词根列表
        top_count = top_words_count
        sub_top_words_list = get_top(lw, top_count, store1, store2, store3, words_already)#,set_top=set_top)
        #print("len(子top词根列表)")
        #print(len(sub_top_words_list))
        for sub_words in sub_top_words_list:
            words_already.add(sub_words)
            sub_idsets = store4[sub_words]
            # 插入子主题
            sub_topic = sub_root_topic.addSubTopic()
            subwords_count = (store3[sub_words].__str__())
            sub_topic.setTitle(sub_words + ' ('+ subwords_count+')')#.decode('utf-8'))
            #print(sub_words)
            #print("$$$$$$$$$$$$\n")
            longtail_words_ids = idsets & sub_idsets
            longtail_words_list = []
            # 打印主题对应的长尾词
            for id in longtail_words_ids:
                longtail_words_list.append(store1[id][0])
            sub_topic_s_kw = sub_topic.addSubTopic()
            sub_topic_s_kw.setTitle(','.join(longtail_words_list[:50]))#.decode('utf-8'))
    #print(workbook.to_prettify_json())
    #xmind.save(workbook)
    # 第2种：只保存思维导图内容content.xml核心文件，适用于没有添加评论、自定义样式和附件的情况
    xmind.save(workbook=workbook, path=dest_file_name)

def cells_to_xmind(cells, outputfile, top_words_count):
    ts = datetime.datetime.now().timestamp() 
    words_already = set() #用来保存已经加入xmind列表的无需再重复加入
    illeagle_words = set() #用来排除一些不适合参与排序词根，下一步可以从文件读取次列表
    illeagle_words_file_path = './dataset/illeagle_words.txt'
    if os.path.isfile(illeagle_words_file_path):
        ws = load_cells_from_file(illeagle_words_file_path, 'utf-8')
        for item in ws:
            illeagle_words.add(item)
            print(item)

    for word in list(illeagle_words):
        words_already.add(word)

    store1,store2,store3,store4 = create_stores(cells, words_already, illeagle_words)
    print('store1 count', len(store1))
    print('store2 count', len(store2))
    print('store3 count', len(store3))
    print('store4 count', len(store4))

    #获取首N个词根
    top_words_list = get_first_topwords(top_words_count, words_already, store3)

    ts2 = datetime.datetime.now().timestamp() 
    gen_my_xmind_file(top_words_list, outputfile, store1, store2, store3, store4, words_already, top_words_count)
    ts3 = datetime.datetime.now().timestamp() 
    print('time passed ', ts2-ts)
    print('create 4 stores time passed (seconds):', ts2-ts)
    print('save to xmind time passed (seconds):', ts3-ts2)

def gs_grouping_sentences_to_xmind(inputfile, outputfile, max_items=10000, encoding='gb18030', topwordscount = 8):
    cells = load_cells_from_file(inputfile, encoding , max_items)
    cells_to_xmind(cells, outputfile, topwordscount)   

def main(argv):
    inputfile = ''
    outputfile = ''
    top_words_count = 8
    encoding = 'gb18030'
    max_items = 1000

    try:
        opts, args = getopt.getopt(argv,"hi:o:c:e:t:",["ifile=","ofile=","encoding=","maxcount="])
    except getopt.GetoptError:
        print('second.py -i <inputfile> -o <outputfile> -c <maxcount 1000 default> -e <encoding gb18030 default> -t <topwordscount 8 default>')
        sys.exit(2)
    for opt, arg in opts:
        if opt == '-h':
            print('second.py -i <inputfile> -o <outputfile> -c <maxcount 1000 default> -e <encoding gb18030 default> -t <topwordscount 8 default>')
            sys.exit()
        elif opt in ("-i", "--ifile"):
            inputfile = arg
        elif opt in ("-o", "--ofile"):
            outputfile = arg
        elif opt in ("-t"):
            top_words_count = int(arg)
        elif opt in ("-e", "--encoding"):
            encoding = (arg)
        elif opt in ("-c", "--maxcount"):
            max_items = int(arg)
    if len(inputfile)==0 or len(outputfile)==0:
        print('second.py -i <inputfile> -o <outputfile> -c <maxcount 1000 default> -e <encoding gb18030 default> -t <topwordscount 8 default>')
        sys.exit(2)

    print('Input file is "', inputfile)
    print('Output file is "', outputfile)
    gs_grouping_sentences_to_xmind(inputfile, outputfile, max_items=10000, encoding='gb18030', topwordscount = 8)


if __name__ == "__main__":
    main(sys.argv[1:])



# -*- coding: utf-8 -*-
"""
:集合各种计算句子间距离的函数，可以自己选择使用
"""
import distance
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.feature_extraction.text import TfidfVectorizer
import numpy as np
from scipy.linalg import norm
import jieba
import numpy as np
import re

def edit_distance(s1, s2):
    return distance.levenshtein(s1, s2)

#jaccard
def jaccard_similarity(s1, s2):
    def add_space(s):
        return ' '.join(list(s))
    
    # 将字中间加入空格
    s1, s2 = add_space(s1), add_space(s2)
    # 转化为TF矩阵
    cv = CountVectorizer(tokenizer=lambda s: s.split())
    corpus = [s1, s2]
    vectors = cv.fit_transform(corpus).toarray()
    # 求交集
    numerator = np.sum(np.min(vectors, axis=0))
    # 求并集
    denominator = np.sum(np.max(vectors, axis=0))
    # 计算杰卡德系数
    return 1.0 * numerator / denominator

def tf_similarity(s1, s2):
    def add_space(s):
        return ' '.join(list(s))
    
    # 将字中间加入空格
    s1, s2 = add_space(s1), add_space(s2)
    # 转化为TF矩阵
    cv = CountVectorizer(tokenizer=lambda s: s.split())
    corpus = [s1, s2]
    vectors = cv.fit_transform(corpus).toarray()
    # 计算TF系数
    return np.dot(vectors[0], vectors[1]) / (norm(vectors[0]) * norm(vectors[1]))


def tfidf_similarity(s1, s2):
    def add_space(s):
        return ' '.join(list(s))
    
    # 将字中间加入空格
    s1, s2 = add_space(s1), add_space(s2)
    # 转化为TF矩阵
    cv = TfidfVectorizer(tokenizer=lambda s: s.split())
    corpus = [s1, s2]
    vectors = cv.fit_transform(corpus).toarray()
    # 计算TF系数
    return np.dot(vectors[0], vectors[1]) / (norm(vectors[0]) * norm(vectors[1]))

vectcos_cache = {}

def get_word_vector(s1,s2):
    """
    :param s1: 句子1
    :param s2: 句子2
    :return: 返回句子的余弦相似度
    """
    # 分词
    if s1 in vectcos_cache:
        list_word1 = vectcos_cache[s1]
    else:
        cut1 = jieba.cut(s1)
        list_word1 = (','.join(cut1)).split(',')
        vectcos_cache[s1] = list_word1

    if s2 in vectcos_cache:
        list_word2 = vectcos_cache[s2]
    else:
        cut2 = jieba.cut(s2)
        list_word2 = (','.join(cut2)).split(',')
        vectcos_cache[s2] = list_word2

    #print(list_word1,'###',list_word2)
    # 列出所有的词,取并集
    key_word = list(set(list_word1 + list_word2))
    # 给定形状和类型的用0填充的矩阵存储向量
    word_vector1 = np.zeros(len(key_word))
    word_vector2 = np.zeros(len(key_word))
 
    # 计算词频
    # 依次确定向量的每个位置的值
    for i in range(len(key_word)):
        # 遍历key_word中每个词在句子中的出现次数
        for j in range(len(list_word1)):
            if key_word[i] == list_word1[j]:
                word_vector1[i] += 1
        for k in range(len(list_word2)):
            if key_word[i] == list_word2[k]:
                word_vector2[i] += 1
 
    # 输出向量
    #print(word_vector1)
    #print(word_vector2)
    return word_vector1, word_vector2
 
def cos_dist(vec1,vec2):
    """
    :param vec1: 向量1
    :param vec2: 向量2
    :return: 返回两个向量的余弦相似度
    """
    dist1=float(np.dot(vec1,vec2)/(np.linalg.norm(vec1)*np.linalg.norm(vec2)))
    return dist1


def vectcos_similarity(s1, s2):
    vec1,vec2=get_word_vector(s1,s2)
    dist1=cos_dist(vec1,vec2)
    return dist1

def get_funcname_by_func_type_id(func_type_id):
    if func_type_id == 0:
        return 'vectcos_similarity'
    elif func_type_id == 1:
        return 'edit_distance'
    elif func_type_id == 2:
        return 'jaccard_similarity'
    elif func_type_id == 3:
        return 'tf_similarity'
    elif func_type_id == 4:
        return 'tfidf_similarity'
    else:
        return 'vectcos_similarity'

def get_similarity_val_by_sentences(s1, s2, type_of_func = 0):
    """
    :param s1: 句子1
    :param s2: 句子2
    :param type_of_func 不同的相似度计算
    :return: 返回句子的相似度
    """
    if type_of_func == 0:
        return vectcos_similarity(s1, s2)
    elif type_of_func == 1:
        return edit_distance(s1, s2)
    elif type_of_func == 2:
        return jaccard_similarity(s1, s2)
    elif type_of_func == 3:
        return tf_similarity(s1, s2)
    elif type_of_func == 4:
        return tfidf_similarity(s1, s2)
    else:
        return vectcos_similarity(s1, s2)

if __name__ == "__main__":
    strings = [
    '你在干什么',
    '你在干啥子',
    '你在做什么',
    '你好啊',
    '我喜欢吃香蕉'
    ]

    target = '你在干啥'
    
    for i in range(5):
        print('=================')
        print('type of func:', get_funcname_by_func_type_id(i))
        print('similarity to ', target)
        for string in strings:
            print(string, get_similarity_val_by_sentences(string, target, i))



from distutils.core import setup
setup(
  name = 'groupingsentences',         # How you named your package folder (MyLib)
  packages = ['groupingsentences'],   # Chose the same as "name"
  version = '0.16',      # Start with a small number and increase it with every change you make
  license='apache-2.0',        # Chose a license from here: https://help.github.com/articles/licensing-a-repository
  description = 'groupingsentences',   # Give a short description about your library
  author = 'Gary App',                   # Type in your name
  author_email = 'garyapphub@gmail.com',      # Type in your E-Mail
  url = 'https://github.com/garyapphub/GroupingSentences.git',   # Provide either the link to your github or to your website
  download_url = 'https://github.com/garyapphub/GroupingSentences/archive/v_01_6.tar.gz',    # I explain this later on
  keywords = ['grouping', 'sentences', 'KEYWORDS'],   # Keywords that define your package best
  install_requires=[            # I get to this in a second
          'validators',
          'jieba',
          'xlrd',
          'xlwt',
          'xmind',
          'sklearn',
          'numpy',
          'distance',
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Developers',      # Define that your audience are developers
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: Apache Software License',   # Again, pick a license
    'Programming Language :: Python :: 3',      #Specify which pyhton versions that you want to support
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
  ],
)